{-# LANGUAGE TupleSections #-}
{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE MagicHash #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE ViewPatterns #-}
module Engine.Experiment
( ExperimentResult(..)
, runExperiment
, remapExperimentResult
) where

import Bottom
import Control.Applicative
import Control.Concurrent
import Control.Exception
import Control.Monad.Reader (ask)
import CoreSyn
import CoreUtils
import DataCon
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Engine.EngineHValue
import Engine.EngineM
import Engine.Eval
import Engine.ExperimentResult
import Engine.IrulanRTError
import FastString
import Format
import GHC.Prim
import Id
import MkCore
import MonadUtils (liftIO)
import Plan
import qualified Data.Foldable as F
import qualified Data.Map as Map
import qualified Data.Traversable as T
import System.Time
import TysPrim (anyPrimTy)
import TysWiredIn (unitDataConId)
import Util.GHC

instance Format ExperimentResult where
  format (Redundant exps) = "# == " ++ format exps
  format Ok              = "."
  format (Errors exc)    = "! " ++ show exc
  format (Bottoms bid)   = "?" ++ format bid
  format (Property p)
    | p                  = "True"
    | otherwise          = "False"

runExperiment :: Experiment -> EngineM RanExperiment
runExperiment experiment = do
  mExistingResult <- lookupExperimentResult experiment
  
  case mExistingResult of
    Just existingResult -> return $ CachedResult existingResult
    Nothing -> do
      env <- ask
      expr <- mkExperimentExpression experiment 
      val <- compile expr

      resultVar <- liftIO $ newEmptyMVar
      evalThreadId <- liftIO $ forkIO (evalWithSeq val >>= ((putMVar resultVar) $!))

      evalResult <- liftIO $ watchMemoryAndTime env resultVar

      liftIO (killThread evalThreadId)
      liftIO getAllocations >>= \allocs -> tellLog ("Allocations: " ++ show allocs)

      let simpleCase result hv = storeExperimentResult experiment result >> return (result, hv)

      (ranExp, mEHVal) <- case evalResult of
        EvalException (recoverIrulanRTError -> Just (Bottom i)) -> simpleCase (Bottoms i) Nothing
        EvalException e  -> simpleCase (Errors e) Nothing
        EvalOk newId val -> do
          if experimentIsProperty experiment
            then simpleCase (Property (unsafeCoerce# (toHValue val))) (Just val)
            else do
              if experimentHasBottom experiment
                then simpleCase Ok (Just val)
                else (maybe (Ok, Just val)) ((,Just val) . Redundant) <$> storeExperimentHValueAndId experiment newId val
      return $ NewResult ranExp mEHVal

watchMemoryAndTime :: EngineEnv -> MVar EvalResult -> IO EvalResult
watchMemoryAndTime env er = do
  startTime <- getClockTime
  let endTime = noTimeDiff { tdPicosec = 1000000 * (fromIntegral (maxExpressionTime env)) } `addToClockTime` startTime 
  baseAllocs <- getAllocations
  let endAllocs = baseAllocs + maxExpressionAllocs env

  let loop = do
              newAllocs <- getAllocations
              newTime   <- getClockTime
              mVal      <- tryTakeMVar er

              let go
                    | Just res  <- mVal
                        = return res
                    | newAllocs > endAllocs 
                        = return (EvalException $! SomeException $! ErrorCall "<allocated too much memory!>")
                    | newTime > endTime
                        = return (EvalException $! SomeException $! ErrorCall "<ran out of time!>")
                    | otherwise = yield >> loop
              go
    
  loop

getOrMkExperimentExpression :: Experiment -> EngineM CoreExpr
getOrMkExperimentExpression exp
  | hasNoBottomIds exp = checkExperimentCache exp 
                       >>= maybe (mkExperimentExpression exp) (return . Var) 
  | otherwise = mkExperimentExpression exp

mkExperimentExpression :: Experiment -> EngineM CoreExpr
mkExperimentExpression experiment = do
  rootApp <- mkExperimentRootExpression
  childApps <- F.toList <$> T.mapM mkArgumentExpression (experimentArgs experiment)
  return $ mkCoreApps rootApp childApps
  where
    mkExperimentRootExpression :: EngineM CoreExpr
    mkExperimentRootExpression
      | IdRoot x <- experimentRoot experiment = return $ Var x
      | DestructRoot { destructCaseOf 
                     , destructConstructor 
                     , destructVarIndex 
                     , destructResultType 
                     } <- experimentRoot experiment = do
      caseOf <- getOrMkExperimentExpression destructCaseOf

      patId <- mkSysLocalM (mkFastString "$irulan") (destructResultType)
      let dcArity = dataConRepArity destructConstructor
      let pattern = map (\x -> if x == destructVarIndex 
                                then patId 
                                else (mkWildBinder anyPrimTy)
                        ) [0..(dcArity - 1)]

      return $ mkWildCase caseOf (exprType caseOf) destructResultType 
                 [ (DataAlt destructConstructor , pattern , Var patId)
                 ]
      {- case destructCaseOf of
       -  destructConstructor [wild, wild, var, wild, wild] -> var
       -}

    mkArgumentExpression :: ExperimentArg -> EngineM CoreExpr
    mkArgumentExpression (ArgExperiment exp)  = getOrMkExperimentExpression exp
    mkArgumentExpression (ArgBottom ty bid)   = mkIrulanRTErrorExpr ty (Bottom bid)
    mkArgumentExpression (ArgType ty)         = return $ Type ty
    mkArgumentExpression ArgUnconstrained     = return $ Var unitDataConId

hasNoBottomIds :: Experiment -> Bool
hasNoBottomIds experiment
  = getAll . F.foldMap hasNoBottomIds' $ (experimentArgs experiment)
  where
    hasNoBottomIds' (ArgExperiment exp)  = All $ hasNoBottomIds exp
    hasNoBottomIds' (ArgBottom _ _)      = All $ False
    hasNoBottomIds' (ArgType _)          = All $ True
    hasNoBottomIds' ArgUnconstrained     = All $ True

    hasNoRootBottomIds :: ExperimentRoot -> All
    hasNoRootBottomIds (IdRoot _) = All $ True
    hasNoRootBottomIds (DestructRoot { destructCaseOf }) 
      = All $ hasNoBottomIds destructCaseOf
