{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE ViewPatterns #-}
module Plan.Experiment 
( Experiment(..)
, ExperimentArg(..)
, ExperimentResult(..)
, ExperimentRoot(..)
, WeakExperiment(..)
, canonicaliseExperiment
) where

import Bottom
import Control.Applicative
import Control.Monad.State
import Control.Monad.Writer
import DataCon
import Data.Maybe
import Data.Sequence (Seq)
import Format
import Name
import Outputable
import qualified Data.Foldable as F
import qualified Data.Sequence as Seq
import qualified Data.Traversable as T
import qualified Util.BiMap as BiMap
import Type
import Unique
import Util.BiMap (BiMap)
import Util.GHC ()
import Var

data Experiment
  = Experiment 
  { experimentRoot       :: !ExperimentRoot 
  , experimentArgs       :: !(Seq ExperimentArg)
  , experimentIsProperty :: !Bool
  , experimentHasBottom  :: !Bool
  }
  deriving (Eq, Ord)


data ExperimentRoot
  = IdRoot !Id
  | DestructRoot 
  { destructCaseOf      :: !Experiment
  , destructConstructor :: !DataCon
  , destructVarIndex    :: !Int
  , destructResultType  :: !Type
  }
  deriving (Eq, Ord)

data ExperimentArg
  = ArgExperiment !Experiment
  | ArgUnconstrained
  | ArgBottom     !Type !BottomId
  | ArgType       !Type
  deriving (Eq, Ord)

data ExperimentResult 
  = Ok 
  | Bottoms !BottomId
  deriving (Eq, Ord, Show)

instance Format Experiment where
  format exp = unwords $ [format (experimentRoot exp)] ++ (F.toList $ format <$> experimentArgs exp)

instance Format ExperimentRoot where
  format (IdRoot x) = show x
  format DestructRoot { destructCaseOf
                      , destructConstructor
                      , destructVarIndex
                      , destructResultType
                      } = "case (" ++ format destructCaseOf ++ ") of " ++ show destructConstructor ++ " -> " ++ show destructVarIndex

instance Format ExperimentArg where
  format (ArgExperiment e) 
    | Seq.null (experimentArgs e) = format e
    | otherwise                   = "(" ++ format e ++ ")"
  format (ArgBottom _ b)          = "?" ++ format b
  format (ArgType t)              = "@" ++ show t
  format ArgUnconstrained         = "*"
      
newtype WeakExperiment     = WeakExperiment     Experiment
newtype WeakExperimentArg  = WeakExperimentArg  ExperimentArg
newtype WeakExperimentRoot = WeakExperimentRoot ExperimentRoot

instance Eq WeakExperiment where
  (WeakExperiment l) == (WeakExperiment r) 
    = WeakExperimentRoot (experimentRoot l) == WeakExperimentRoot (experimentRoot r) &&
      and (zipWith (==) (F.toList $ WeakExperimentArg <$> experimentArgs l)
                        (F.toList $ WeakExperimentArg <$> experimentArgs r))

instance Ord WeakExperiment where
  (WeakExperiment l) `compare` (WeakExperiment r) = mkRep l `compare` mkRep r
    where
      mkRep :: Experiment -> (WeakExperimentRoot, Seq WeakExperimentArg)
      mkRep exp = (WeakExperimentRoot . experimentRoot $ exp, WeakExperimentArg <$> experimentArgs exp)

instance Eq WeakExperimentRoot where
  (WeakExperimentRoot (IdRoot l)) == (WeakExperimentRoot (IdRoot r)) = l == r
  (WeakExperimentRoot (l@DestructRoot {})) == (WeakExperimentRoot (r@DestructRoot {}))
    = and [ destructConstructor l == destructConstructor r
          , destructVarIndex l    == destructVarIndex r 
          , WeakExperiment (destructCaseOf l) == WeakExperiment (destructCaseOf r)
          ] 
  _ == _ = False

instance Ord WeakExperimentRoot where
  (WeakExperimentRoot l) `compare` (WeakExperimentRoot r) = mkRep l `compare` mkRep r
    where
      mkRep :: ExperimentRoot -> Either Id (DataCon, Int, WeakExperiment)
      mkRep (IdRoot x) = Left x
      mkRep (d@DestructRoot {}) = Right (destructConstructor d, destructVarIndex d, WeakExperiment $ destructCaseOf d)

instance Eq WeakExperimentArg where
  (WeakExperimentArg la) == (WeakExperimentArg ra) = la =~= ra
    where
      ArgExperiment (WeakExperiment -> l) =~= ArgExperiment (WeakExperiment -> r) = l == r
      ArgUnconstrained =~= ArgUnconstrained = True
      ArgBottom _ l =~= ArgBottom _ r       = l == r
      ArgType _ =~= ArgType _               = True
      _ =~= _                               = False

instance Ord WeakExperimentArg where
  (WeakExperimentArg l) `compare` (WeakExperimentArg r) = mkRep l `compare` mkRep r
    where
      mkRep :: ExperimentArg -> Either WeakExperiment (Either Bool BottomId)
      mkRep (ArgExperiment sub) = Left (WeakExperiment sub)
      mkRep ArgUnconstrained    = Right (Left False)
      mkRep (ArgBottom _ bid)   = Right (Right bid)
      mkRep (ArgType _)         = Right (Left True)

canonicaliseExperiment :: Experiment -> (Experiment, BiMap BottomId BottomId)
canonicaliseExperiment = flip evalState initialBottomId . runWriterT . canonicaliseExperiment'
  where
    canonicaliseExperiment' exp = do
      root <- canonicaliseRoot (experimentRoot exp)
      args <- T.mapM canonicaliseArg (experimentArgs exp)
      return $ exp { experimentRoot = root
                   , experimentArgs = args 
                   }

    canonicaliseArg (ArgExperiment exp) = ArgExperiment <$> canonicaliseExperiment' exp
    canonicaliseArg ArgUnconstrained    = return ArgUnconstrained
    canonicaliseArg (ArgBottom t obid)  = do
      nbid <- lift nextBottomId
      tell (BiMap.singleton obid nbid)
      return $ ArgBottom t nbid
    canonicaliseArg t@(ArgType _) = return t

    canonicaliseRoot i@(IdRoot _) = return i
    canonicaliseRoot d@(DestructRoot { destructCaseOf }) = do
      destructCaseOf' <- canonicaliseExperiment' destructCaseOf
      return $ d { destructCaseOf = destructCaseOf'
                 }

    nextBottomId = State (\id -> (id, succ id))
