{-# LANGUAGE NamedFieldPuns #-}
{-# LANGUAGE PatternGuards #-}
{-# LANGUAGE BangPatterns #-}
module Plan.TreePlan
( TreePlan(..)
, generateTreePlan
) where

import Bottom
import Control.Applicative
import Control.Monad
import DataCon
import Data.List
import Data.Map (Map)
import Data.Maybe
import Data.Monoid
import Data.Ord
import Data.Sequence (Seq)
import Data.Set (Set)
import Format
import IdDB
import Outputable hiding (trace)
import Plan.Experiment
import Plan.IExperiment
import Plan.PlanM
import qualified Data.Map as Map
import qualified Data.Sequence as Seq
import qualified Data.Set as Set
import qualified Data.Traversable as T
import qualified Util.UniqMap as UniqMap
import qualified Util.UniqSet as UniqSet
import Type
import TysWiredIn (unitTy)
import Util
import Util.GHC
import Util.UniqMap (UniqMap)
import Util.UniqSet (UniqSet)
import Var
import VarEnv

data TreePlan
  = TPStep
  { tpExperiment  :: !Experiment
  , tpOnOk        :: !(Either TreePlan (UniqMap DataCon (Seq TreePlan)))
  , tpOnBottom    :: !(Map BottomId (Seq TreePlan))
  }
  | TPSplit
  { tpChildren    :: !(Seq TreePlan)
  }

stop :: TreePlan
stop = TPSplit { tpChildren = Seq.empty }

data PlanContext
  = PlanContext
  { instantiate   :: !(IExperiment -> TvSubst -> InScopeSet -> PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan))))
  , targetType    :: !Type
  , inScopeTyVars :: !InScopeSet
  , usedIds       :: !(Set Id)
  }

generateTreePlan :: IdDB -> TreePlan
generateTreePlan !iddb = generateTreePlan' ids
  where
    generateTreePlan' = TPSplit . Seq.fromList 
                      . runPlanM iddb . mapM buildAPlan
    ids = functionsToTest iddb
    
    buildAPlan id = buildTreePlan (makeInitialExperiment id emptyTvSubst) newContext
      where
        newContext = PlanContext { instantiate   = tryToDestructTreePlan
                                 , targetType    = snd . splitFunTys $ targetRawType
                                 , inScopeTyVars = extendInScopeSetList emptyInScopeSet targetTyVars
                                 , usedIds       = Set.singleton id
                                 }

        (targetTyVars, targetRawType) = splitForAllTys . varType $ id
    

buildTreePlan :: IExperiment -> PlanContext -> PlanM TreePlan
buildTreePlan !exp (dat@PlanContext { instantiate   = instantiate
                                    , targetType    = targetType
                                    , inScopeTyVars = inScopeTyVars
                                    , usedIds       = usedIds
                                    })
  = do
    mExpRes <- lookupExperimentResult exp

    let uncachedVersion 
          = do
              okPlan      <- withExperimentResult exp Ok mkOkPlan
              bottomPlans <- T.sequence $ Map.mapWithKey (\bid -> withExperimentResult exp (Bottoms bid)) (bottomPlanMs dat exp)

              return $! TPStep { tpExperiment = iExperimentToExperiment exp
                               , tpOnOk       = okPlan
                               , tpOnBottom   = bottomPlans
                               }
    case mExpRes of
      Just Ok                 -> do
        okP <- mkOkPlan
        case okP of
          Left p  -> return p
          Right _ -> trace ("Cached ok, but untraced: " ++ format exp) uncachedVersion
      Just (Bottoms bid)      -> do
        kids <- fromMaybe (return Seq.empty) $ Map.lookup bid (bottomPlanMs dat exp)
        return $! TPSplit { tpChildren = kids }
      Nothing -> do
        uncachedVersion

  where
    mkOkPlan :: PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan)))
    mkOkPlan
      | Just match <- maybeMatch = instantiate exp match inScopeTyVars
      -- | trace ("No match: " ++ show expType ++ " vs " ++ show targetType) False = undefined
      | otherwise                = withNewBottomId $ \bid -> let exp' = addBottomArg exp bid
                                                              in Left <$> buildTreePlan exp' dat
      where
        maybeMatch = ut
        expType = iExperimentType exp
        ut = unifyTypes expType targetType
    
bottomPlanMs :: PlanContext -> IExperiment -> Map BottomId (PlanM (Seq TreePlan))
bottomPlanMs (dat@PlanContext { inScopeTyVars, usedIds, targetType }) exp = Map.mapWithKey makePlansForAType (bottomIdTypes exp)
  where
    makePlansForAType :: BottomId -> Type -> PlanM (Seq TreePlan)
    makePlansForAType !bid !ty
      | isTyVarTy ty = Seq.singleton <$> buildTreePlan (substBottomId bid (IArgUnconstrained (getTyVar "Impossible!" ty) bid) exp) dat
    makePlansForAType !bid !ty = do
      possibles <- Seq.fromList . sortBy (comparing (idNumArguments . fst3)) <$> askIdsForType inScopeTyVars ty
      T.sequence $ processSource <$> possibles
      where
        processSource :: (Id, TvSubst, InScopeSet) -> PlanM TreePlan
        processSource (source, match, iss)
          | source `Set.member` usedIds = return stop
          | otherwise = buildTreePlan (makeInitialExperiment source match) 
                                      dat { instantiate = instL newInstantiate
                                          , targetType = ty
                                          , inScopeTyVars = iss
                                          , usedIds = Set.insert source usedIds
                                          }
          where
            newInstantiate newExp newSubst newTyVars
              = iExperimentToIExperimentArgument newExp $ \ta ->
                  let substExp = substExpFn newSubst
                               . substBottomId bid ta
                               $ exp
                  in buildTreePlan substExp dat { targetType    = substTy newSubst targetType
                                                , inScopeTyVars = newTyVars
                                                }


instL :: (IExperiment -> TvSubst -> InScopeSet -> PlanM TreePlan) 
      ->  IExperiment -> TvSubst -> InScopeSet -> PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan)))
instL !f !a1 !a2 !a3 = Left <$> f a1 a2 a3

instR :: (IExperiment -> TvSubst -> InScopeSet -> PlanM (UniqMap DataCon (Seq TreePlan)))
      ->  IExperiment -> TvSubst -> InScopeSet -> PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan)))
instR !f !a1 !a2 !a3 = Right <$> f a1 a2 a3

tryToDestructTreePlan :: IExperiment -> TvSubst -> InScopeSet -> PlanM (Either TreePlan (UniqMap DataCon (Seq TreePlan)))
tryToDestructTreePlan iexp subst iss
  | (argTys, resultTy) <- splitFunTys ty
  , not . null $ argTys                                
    = Left <$> buildTreePlan iexp (PlanContext { instantiate = tryToDestructTreePlan
                                               , targetType = resultTy
                                               , inScopeTyVars = iss
                                               , usedIds = Set.empty
                                               })
  where
    ty = iExperimentType iexp

tryToDestructTreePlan iexp subst iss
  | Just (tyCon, tyArgs)     <- splitTyConApp_maybe ty = do
    dcs <- UniqSet.toList <$> askConstructorsForTyCon tyCon
    theMap <- mconcat <$> mapM (mkDataConMap tyArgs) dcs
    return $! Right $! theMap
  where
    ty = iExperimentType iexp
    mkDataConMap :: [Type] -> DataCon -> PlanM (UniqMap DataCon (Seq TreePlan))
    mkDataConMap tyArgs dc = UniqMap.singleton dc <$> destructDataCon tyArgs dc

    destructDataCon :: [Type] -> DataCon -> PlanM (Seq TreePlan)
    destructDataCon tyArgs dc = Seq.fromList <$> (sequence . catMaybes . zipWith mkOne [0..] . map (substTy subst) . dataConRepArgTys $ dc)
      where
        subst = zipOpenTvSubst (dataConUnivTyVars dc) tyArgs

        mkOne :: Int -> Type -> Maybe (PlanM TreePlan)
        mkOne i t
          | isUnLiftedType t = Nothing
          | otherwise = Just $ do
            onOk <- withExperimentResult theExperiment Ok (tryToDestructTreePlan theExperiment subst iss)
            let dat = PlanContext { instantiate = tryToDestructTreePlan
                                  , targetType = t
                                  , inScopeTyVars = iss
                                  , usedIds = Set.empty 
                                  }
            bottomPlans <- T.sequence $ Map.mapWithKey (\bid -> withExperimentResult theExperiment (Bottoms bid)) (bottomPlanMs dat theExperiment)
            return $ TPStep { tpExperiment = iExperimentToExperiment theExperiment
                            , tpOnOk       = onOk
                            , tpOnBottom   = bottomPlans
                            }
            where
              theExperiment = destructIExperiment iexp dc i t

tryToDestructTreePlan _ _ _ = return $! Left $! stop
